#define CLOCK A4
#define STROBE A3
#define DATA A5
#define PARITY A2

#define KEYS_A 8
#define KEYS_B 7
#define KEYS_C 6
#define KEYS_1 5
#define KEYS_2 4
#define KEYS_3 3
#define KEYS_4 2

enum counterType {
  LOWSIDE,
  HIGHSIDE
};

counterType curCounter = LOWSIDE;

byte DIGITS_SET[10] = {0b11111100, 0b01100000, 0b11011010, 0b11110010, 0b01100110, 0b10110110, 0b10111110, 0b11100000, 0b11111110, 0b11110110};
byte curDigits[14] = {0};

void clockPulse(int duration) {
  /*
     Sends a clock pulse with the specified duration
  */
  digitalWrite(CLOCK, HIGH);
  delay(duration);
  digitalWrite(CLOCK, LOW);
}

void shiftData(unsigned int data) {
  /*
     Shifts out two data bytes to the shift register

     High Byte (Low-side segment driver mask) bit order:
     7 6 5 4 3 2 1 0
     a b c d e f g

     Low Byte High Nibble: Low-side digit driver address
     Low Byte Low Nibble: High-side digit driver address
  */
  digitalWrite(STROBE, HIGH);
  for (int i = 15; i >= 0; i--) {
    digitalWrite(DATA, (data >> i) & 1);
    clockPulse(0);
  }
}

void selectCounter(counterType cnt) {
  /*
     Enables the selected counter
  */
  if (cnt == LOWSIDE) {
    digitalWrite(DATA, HIGH);
  } else if (cnt == HIGHSIDE) {
    digitalWrite(DATA, LOW);
  }
  digitalWrite(STROBE, LOW);
  curCounter = cnt;
}

void iterateCounter(byte selectedSegments) {
  /*
     Iterates the currently selected counter,
     skipping non-selected segments
     by not pausing as long
     (only needed for HIGHSIDE counter)

     selectedSegments bit order:
     7 6 5 4 3 2 1 0
       a f b g e c d
  */
  for (int i = 0; i < 8; i++) {
    if (curCounter == LOWSIDE || (curCounter == HIGHSIDE && (selectedSegments >> i) & 1)) {
      clockPulse(1);
    } else {
      clockPulse(0);
    }
  }
}

void setDigit(byte digit, byte value) {
  // Reset
  shiftData(((digit & 0x0f) << 4));
  selectCounter(HIGHSIDE);
  iterateCounter(0xff);

  // Set
  shiftData((value << 8) | (digit & 0x0f));
  selectCounter(LOWSIDE);
  iterateCounter(0);
}

byte getKey() {
  byte key = 0xff;
  digitalWrite(KEYS_A, LOW);
  if (!digitalRead(KEYS_1)) key = 1;
  if (!digitalRead(KEYS_2)) key = 4;
  if (!digitalRead(KEYS_3)) key = 7;
  if (!digitalRead(KEYS_4)) key = 10;
  digitalWrite(KEYS_A, HIGH);
  digitalWrite(KEYS_B, LOW);
  if (!digitalRead(KEYS_1)) key = 2;
  if (!digitalRead(KEYS_2)) key = 5;
  if (!digitalRead(KEYS_3)) key = 8;
  if (!digitalRead(KEYS_4)) key = 0;
  digitalWrite(KEYS_B, HIGH);
  digitalWrite(KEYS_C, LOW);
  if (!digitalRead(KEYS_1)) key = 3;
  if (!digitalRead(KEYS_2)) key = 6;
  if (!digitalRead(KEYS_3)) key = 9;
  if (!digitalRead(KEYS_4)) key = 11;
  digitalWrite(KEYS_C, HIGH);
  return key;
}

byte waitForKey() {
  byte key;
  while (true) {
    key = getKey();
    if (key != 0xff) break;
  }
  while (getKey() != 0xff);
  return key;
}

void setup() {
  pinMode(CLOCK, OUTPUT);
  pinMode(STROBE, OUTPUT);
  pinMode(DATA, OUTPUT);
  pinMode(PARITY, INPUT);

  pinMode(KEYS_A, OUTPUT);
  pinMode(KEYS_B, OUTPUT);
  pinMode(KEYS_C, OUTPUT);
  pinMode(KEYS_1, INPUT_PULLUP);
  pinMode(KEYS_2, INPUT_PULLUP);
  pinMode(KEYS_3, INPUT_PULLUP);
  pinMode(KEYS_4, INPUT_PULLUP);

  for (int i = 0; i < 15; i++) {
    setDigit(i + 1, DIGITS_SET[8]);
  }
  delay(2000);
  for (int i = 0; i < 14; i++) {
    setDigit(i + 1, 0);
  }
}

byte digit = 1;
void loop() {
  setDigit(digit, 0b00010000);
  byte key = waitForKey();
  if (key < 10) {
    curDigits[digit - 1] = DIGITS_SET[key];
    setDigit(digit, curDigits[digit - 1]);
    digit++;
  } else {
    if (key == 10) {
      setDigit(digit, curDigits[digit - 1]);
      digit--;
    } else if (key == 11) {
      setDigit(digit, curDigits[digit - 1]);
      digit++;
    }
  }
  if (digit < 1) digit = 15;
  if (digit > 15) digit = 1;
}
